-- fifo
--
-- fifo realisiert eine fifo-Warteschlange nach Uebung 2, (Entwurf digitaler Systeme)
-- Autor: Andreas Gutscher
-- Datum: 5.6.2000

library ieee;
use ieee.std_logic_1164.all;
library work;
use work.all;

entity fifo is
	generic(
		WIDTH : positive := 4;
		MAX_ITEMS : positive := 3);
	port(
		clk : in std_logic; -- Takt
		-- Eingangsseite:
		resetn : in std_logic; -- negierter asynchroner Reset-Eingang
		wr_ena : in std_logic; -- Enable fr das Einschreiben eines Datums
		data_in : in std_logic_vector(WIDTH-1 downto 0); -- Eingang
		full : out std_logic; -- FIFO ist voll
		-- Ausgangsseite:
		data_out : out std_logic_vector(WIDTH-1 downto 0); -- Ausgang
		empty : out std_logic; -- FIFO ist leer
		rd_ena : in std_logic); -- Datum am Ausgang wurde ausgelesen
end entity fifo;


architecture struct of fifo is
	type data_array is array (max_items downto 0) of std_logic_vector(width-1 downto 0);
	signal select_cell : std_logic_vector(max_items downto 1);
	signal enable_cell : std_logic_vector(max_items downto 1);
	signal d : data_array;
begin

	Zellen: for cell in max_items downto 1 generate 
	begin
	
		Zelle: entity fifo_cell
			generic map(width)
			port map(
				clk => clk,
				clr_n => '1',
				d_in_0 => data_in,
				d_in_1 => d(cell),
				d_out => d(cell-1),
				sel => select_cell(cell),
				ena => enable_cell(cell)
			);
			
	end generate Zellen;

	d(max_items) <= data_in;
	data_out <= d(0);

	Steuerung: entity fifo_ctrl
		generic map(width, max_items)
		port map(
			clk => clk,
			clr_n => resetn,
			wr_ena => wr_ena,
			rd_ena => rd_ena,
			full => full,
			empty => empty,
			sel => select_cell,
			ena => enable_cell
		);

end architecture struct;


architecture behave of fifo is
begin
	process (clk, resetn, wr_ena, data_in, rd_ena) is
		subtype Anzahl_items is natural range 0 to MAX_ITEMS;
		subtype Speicherstelle is natural range 1 to MAX_ITEMS;
		variable items : Anzahl_items := 0; -- Anzahl der Daten im FIFO
		type FIFO_Speicher is array (Speicherstelle) of std_logic_vector(WIDTH-1 downto 0);
		variable data : FIFO_Speicher;
		variable first : Speicherstelle;
	begin

		assert not (items=MAX_ITEMS and (wr_ena='1' or wr_ena='H')) report "Sie haben versucht, ein Datum in die FIFO zu schreiben (wr_ena='1') obwohl die FIFO voll ist (full='1') !" severity error; 
		assert not (items=0 and (rd_ena='1' or rd_ena='H')) report "Sie haben ein Datum aus der FIFO gelesen (rd_ena='1') obwohl die FIFO leer ist (empty='1') !" severity error; 

		if (resetn='0' or resetn='L') then 
			items :=0; 
		elsif clk='1' and clk'event then -- positive Taktflanke
		
			if (rd_ena='1' or rd_ena='H') and items > 0 then -- ein Element wurde ausgelesen
				first := first mod MAX_ITEMS + 1; -- first auf das naechste Datum setzen 
				items := items - 1; 
			end if;
			
			if (wr_ena='1' or wr_ena='H') and items < MAX_ITEMS then -- Datum in FIFO-Kette schreiben
				data((first + items -1) mod MAX_ITEMS +1) := data_in; 
				items := items + 1;
			end if;
			
		end if;

		if items > 0 then
			data_out <= data(first);
		end if;

		if items = 0 then
			empty <= '1';
			full <= '0';
		elsif items = MAX_ITEMS then
			full <= '1';
			empty <= '0';
		else
			full <= '0';
			empty <= '0';
		end if;
		
	end process;
end architecture behave;




